import sys
import itertools

from google.protobuf.compiler.plugin_pb2 import CodeGeneratorRequest
from google.protobuf.compiler.plugin_pb2 import CodeGeneratorResponse
from google.protobuf import descriptor_pb2
from purerpc import Cardinality


def generate_import_statement(proto_name):
    module_path = proto_name[:-len(".proto")].replace("-", "_").replace("/", ".") + "_pb2"
    alias = get_python_module_alias(proto_name)
    return "import " + module_path + " as " + alias


def get_python_module_alias(proto_name):
    package_name = proto_name[:-len(".proto")]
    return package_name.replace("/", "_dot_").replace("-", "_") + "__pb2"


def simple_type(type_):
    simple_type = type_.split(".")[-1]
    return simple_type


def get_python_type(proto_name, proto_type):
    if proto_type.startswith("."):
        return get_python_module_alias(proto_name) + "." + simple_type(proto_type)
    else:
        return proto_type


def generate_single_proto(proto_file: descriptor_pb2.FileDescriptorProto,
                          proto_for_entity):
    lines = ["import purerpc"]
    lines.append(generate_import_statement(proto_file.name))
    for dep_module in proto_file.dependency:
        lines.append(generate_import_statement(dep_module))
    for service in proto_file.service:
        if proto_file.package:
            fully_qualified_service_name = proto_file.package + "." + service.name
        else:
            fully_qualified_service_name = service.name

        lines.append("\n\nclass {service_name}Servicer(purerpc.Servicer):".format(service_name=service.name))
        for method in service.method:
            plural_suffix = "s" if method.client_streaming else ""
            fmt_string = ("    async def {method_name}(self, input_message{plural_suffix}):\n"
                          "        raise NotImplementedError()\n")
            lines.append(fmt_string.format(
                method_name=method.name,
                plural_suffix=plural_suffix,
            ))
        fmt_string = ("    @property\n"
                      "    def service(self) -> purerpc.Service:\n"
                      "        service_obj = purerpc.Service(\n"
                      "            \"{fully_qualified_service_name}\"\n"
                      "        )")
        lines.append(fmt_string.format(fully_qualified_service_name=fully_qualified_service_name))
        for method in service.method:
            input_proto = proto_for_entity[method.input_type]
            output_proto = proto_for_entity[method.output_type]
            cardinality = Cardinality.get_cardinality_for(request_stream=method.client_streaming,
                                                          response_stream=method.server_streaming)
            fmt_string = ("        service_obj.add_method(\n"
                          "            \"{method_name}\",\n"
                          "            self.{method_name},\n"
                          "            purerpc.RPCSignature(\n"
                          "                purerpc.{cardinality},\n"
                          "                {input_type},\n"
                          "                {output_type},\n"
                          "            )\n"
                          "        )")
            lines.append(fmt_string.format(
                method_name=method.name,
                cardinality=cardinality,
                input_type=get_python_type(input_proto, method.input_type),
                output_type=get_python_type(output_proto, method.output_type),
            ))
        lines.append("        return service_obj\n\n")

        fmt_string = ("class {service_name}Stub:\n"
                      "    def __init__(self, channel):\n"
                      "        self._client = purerpc.Client(\n"
                      "            \"{fully_qualified_service_name}\",\n"
                      "            channel\n"
                      "        )")
        lines.append(fmt_string.format(
            service_name=service.name,
            fully_qualified_service_name=fully_qualified_service_name
        ))
        for method in service.method:
            input_proto = proto_for_entity[method.input_type]
            output_proto = proto_for_entity[method.output_type]
            cardinality = Cardinality.get_cardinality_for(request_stream=method.client_streaming,
                                                          response_stream=method.server_streaming)
            fmt_string = ("        self.{method_name} = self._client.get_method_stub(\n"
                          "            \"{method_name}\",\n"
                          "            purerpc.RPCSignature(\n"
                          "                purerpc.{cardinality},\n"
                          "                {input_type},\n"
                          "                {output_type},\n"
                          "            )\n"
                          "        )")
            lines.append(fmt_string.format(
                method_name=method.name,
                cardinality=cardinality,
                input_type=get_python_type(input_proto, method.input_type),
                output_type=get_python_type(output_proto, method.output_type),
            ))

    return "\n".join(lines)


def main():
    request = CodeGeneratorRequest.FromString(sys.stdin.buffer.read())

    files_to_generate = set(request.file_to_generate)

    response = CodeGeneratorResponse()
    proto_for_entity = dict()
    for proto_file in request.proto_file:
        package_name = proto_file.package
        for named_entity in itertools.chain(proto_file.message_type, proto_file.enum_type,
                                            proto_file.service, proto_file.extension):
            if package_name:
                fully_qualified_name = ".".join(["", package_name, named_entity.name])
            else:
                fully_qualified_name = "." + named_entity.name
            proto_for_entity[fully_qualified_name] = proto_file.name
    for proto_file in request.proto_file:
        if proto_file.name in files_to_generate:
            out = response.file.add()
            out.name = proto_file.name.replace('-', "_").replace('.proto', "_grpc.py")
            out.content = generate_single_proto(proto_file, proto_for_entity)
    sys.stdout.buffer.write(response.SerializeToString())



import logging
import math
import os
import platform
import pdb

import anyio

_log = logging.getLogger(__name__)


def is_linux():
    return platform.system() == "Linux"


def is_darwin():
    return platform.system() == "Darwin"


def is_windows():
    return platform.system() == "Windows"


def get_linux_kernel_version():
    if not is_linux():
        return None
    release = platform.release()
    if not release:
        return None
    return tuple(map(int, release.split("-")[0].split(".")))


async def print_memory_growth_statistics(interval_sec=10.0, set_pdb_trace_every=math.inf):
    num_iters = 0
    import objgraph
    while True:
        num_iters += 1
        await anyio.sleep(interval_sec)
        objgraph.show_growth()
        if num_iters == set_pdb_trace_every:
            pdb.set_trace()
            num_iters = 0


def run(func, *args, backend=None, backend_options=None):
    """wrapper for anyio.run() with some purerpc-specific conventions

      * if `backend` is None, read it from PURERPC_BACKEND environment variable,
        (still defaulting to asyncio)
      * allow "uvloop" as a value of `backend` (normally uvloop needs to
        be specified via `backend_options` under asyncio)
      * if uvloop is selected, raise ModuleNotFoundError if uvloop isn't installed
    """

    if backend is None:
        backend = os.getenv("PURERPC_BACKEND", "asyncio")
    _log.info("purerpc.run() selected {} backend".format(backend))
    if backend == "uvloop":
        backend = "asyncio"
        options = dict(use_uvloop=True)
        if backend_options is None:
            backend_options = options
        else:
            backend_options.update(options)
    if backend == "asyncio" and backend_options and backend_options.get('use_uvloop'):
        # Since anyio.run() will silently fall back when uvloop isn't available,
        # make the requirement explicit.
        import uvloop
    return anyio.run(func, *args, backend=backend, backend_options=backend_options)



import urllib.parse
import datetime

from .headers import HeaderDict
from .exceptions import ProtocolError
from .status import Status


class Event:
    pass


class WindowUpdated(Event):
    def __init__(self, stream_id, delta):
        self.stream_id = stream_id
        self.delta = delta

    def __repr__(self):
        return "<WindowUpdated stream_id:%s, delta:%s>" % (
            self.stream_id, self.delta
        )



class RequestReceived(Event):
    def __init__(self, stream_id: int, scheme: str, service_name: str, method_name: str,
                 content_type: str):
        self.stream_id = stream_id
        self.scheme = scheme
        self.service_name = service_name
        self.method_name = method_name
        self.content_type = content_type
        self.authority = None
        self.timeout = None
        self.message_type = None
        self.message_encoding = None
        self.message_accept_encoding = None
        self.user_agent = None
        self.custom_metadata = ()

    @staticmethod
    def parse_from_stream_id_and_headers_destructive(stream_id: int, headers: HeaderDict):
        if headers.pop(":method") != "POST":
            raise ProtocolError("Unsupported method {}".format(headers[":method"]))

        scheme = headers.pop(":scheme")
        if scheme not in ["http", "https"]:
            raise ProtocolError("Scheme should be either http or https")

        if headers[":path"].startswith("/"):
            service_name, method_name = headers.pop(":path")[1:].split("/")
        else:
            raise ProtocolError("Path should be /<service_name>/<method_name>")

        if "te" not in headers or headers["te"] != "trailers":
            raise ProtocolError("te header not found or not equal to 'trailers', "
                                "using incompatible proxy?")
        else:
            headers.pop("te")

        content_type = headers.pop("content-type")
        if not content_type.startswith("application/grpc"):
            raise ProtocolError("Content type should start with application/grpc")

        event = RequestReceived(stream_id, scheme, service_name, method_name, content_type)

        if ":authority" in headers:
            event.authority = headers.pop(":authority")

        if "grpc-timeout" in headers:
            timeout_string = headers.pop("grpc-timeout")
            timeout_value, timeout_unit = int(timeout_string[:-1]), timeout_string[-1:]
            if timeout_unit == "H":
                event.timeout = datetime.timedelta(hours=timeout_value)
            elif timeout_unit == "M":
                event.timeout = datetime.timedelta(minutes=timeout_value)
            elif timeout_unit == "S":
                event.timeout = datetime.timedelta(seconds=timeout_value)
            elif timeout_unit == "m":
                event.timeout = datetime.timedelta(milliseconds=timeout_value)
            elif timeout_unit == "u":
                event.timeout = datetime.timedelta(microseconds=timeout_value)
            elif timeout_unit == "n":
                event.timeout = datetime.timedelta(microseconds=timeout_value / 1000)
            else:
                raise ProtocolError("Unknown timeout unit: {}".format(timeout_unit))

        if "grpc-encoding" in headers:
            event.message_encoding = headers.pop("grpc-encoding")

        if "grpc-accept-encoding" in headers:
            event.message_accept_encoding = headers.pop("grpc-accept-encoding").split(",")

        if "user-agent" in headers:
            event.user_agent = headers.pop("user-agent")

        if "grpc-message-type" in headers:
            event.message_type = headers.pop("grpc-message-type")

        event.custom_metadata = tuple(header for header_name in list(headers.keys())
                                      for header in headers.extract_headers(header_name))
        return event

    def __repr__(self):
        fmt_string = ("<purerpc.grpclib.events.RequestReceived stream_id: {stream_id}, "
                      "service_name: {service_name}, method_name: {method_name}>")
        return fmt_string.format(
            stream_id=self.stream_id,
            service_name=self.service_name,
            method_name=self.method_name,
        )


class MessageReceived(Event):
    def __init__(self, stream_id: int, data: bytes, flow_controlled_length: int):
        self.stream_id = stream_id
        self.data = data
        self.flow_controlled_length = flow_controlled_length

    def __repr__(self):
        fmt_string= ("<purerpc.grpclib.events.MessageReceived stream_id: {stream_id}, "
                     "flow_controlled_length: {flow_controlled_length}>")
        return fmt_string.format(
            stream_id=self.stream_id,
            flow_controlled_length=self.flow_controlled_length,
        )


class RequestEnded(Event):
    def __init__(self, stream_id: int):
        self.stream_id = stream_id

    def __repr__(self):
        fmt_string = "<purerpc.grpclib.events.RequestEnded stream_id: {stream_id}>"
        return fmt_string.format(
            stream_id=self.stream_id,
        )


class ResponseReceived(Event):
    def __init__(self, stream_id: int, content_type: str):
        self.stream_id = stream_id
        self.content_type = content_type
        self.message_encoding = None
        self.message_accept_encoding = None
        self.custom_metadata = ()

    @staticmethod
    def parse_from_stream_id_and_headers_destructive(stream_id: int, headers: HeaderDict):
        if int(headers.pop(":status")) != 200:
            raise ProtocolError("http status is not 200")

        content_type = headers.pop("content-type")
        if not content_type.startswith("application/grpc"):
            raise ProtocolError("Content type should start with application/grpc")

        event = ResponseReceived(stream_id, content_type)

        if "grpc-encoding" in headers:
            event.message_encoding = headers.pop("grpc-encoding")

        if "grpc-accept-encoding" in headers:
            event.message_accept_encoding = headers.pop("grpc-accept-encoding").split(",")

        event.custom_metadata = tuple(header for header_name in list(headers.keys())
                                      for header in headers.extract_headers(header_name))
        return event

    def __repr__(self):
        fmt_string = "<purerpc.grpclib.events.ResponseReceived stream_id: {stream_id} content_type: {content_type}>"
        return fmt_string.format(
            stream_id=self.stream_id,
            content_type=self.content_type,
        )


class ResponseEnded(Event):
    def __init__(self, stream_id: int, status: Status):
        self.stream_id = stream_id
        self.status = status
        self.custom_metadata = ()

    @staticmethod
    def parse_from_stream_id_and_headers_destructive(stream_id: int, headers: HeaderDict):
        if "grpc-status" not in headers:
            raise ProtocolError("Expected grpc-status in trailers")

        status_code = int(headers.pop("grpc-status"))
        if "grpc-message" in headers:
            status_message = urllib.parse.unquote(headers.pop("grpc-message"))
        else:
            status_message = ""

        event = ResponseEnded(stream_id, Status(status_code, status_message))

        event.custom_metadata = tuple(header for header_name in list(headers.keys())
                                      for header in headers.extract_headers(header_name))
        return event

    def __repr__(self):
        fmt_string = "<purerpc.grpclib.events.ResponseEnded stream_id: {stream_id}, status: {status}>"
        return fmt_string.format(
            stream_id=self.stream_id,
            status=self.status,
        )



import h2.connection
import h2.stream


def new_h2_connection_data_to_send(self, amount=None):
    if amount is None:
        data = bytes(self._data_to_send)
        self._data_to_send = bytearray()
        return data
    else:
        data = bytes(self._data_to_send[:amount])
        self._data_to_send = self._data_to_send[amount:]
        return data


def new_h2_connection_clear_outbound_data_buffer(self):
    self._data_to_send = bytearray()


def monkey_patch_h2_data_to_send():
    h2.connection.H2Connection.data_to_send = new_h2_connection_data_to_send
    h2.connection.H2Connection.clear_outbound_data_buffer = new_h2_connection_clear_outbound_data_buffer


def monkey_patch_h2_reset_on_closed_stream_bug():
    """
    H2 raises h2.exceptions.StreamClosedError when receiving RST_STREAM on a closed stream that was
    not closed by endpoint, for example:

    -> HEADERS frame
    -> DATA frame with END_STREAM flag set  # stream is now half-closed (local)
    <- HEADERS frame
    <- DATA frame
    <- HEADERS frame with END_STREAM flag set  # stream is now closed
    <- RST_STREAM

    This is the case when calling unary-unary requests with purerpc client and grpcio server
    (it sends END_STREAM + RST_STREAM because the RPC call is unary-unary and it cannot receive
    any more data).

    To fix this, we use a workaround.
    """
    h2.stream._transitions[h2.stream.StreamState.CLOSED,
                           h2.stream.StreamInputs.RECV_RST_STREAM] = \
        (None, h2.stream.StreamState.CLOSED)

    h2.stream._transitions[h2.stream.StreamState.CLOSED,
                           h2.stream.StreamInputs.RECV_WINDOW_UPDATE] = \
        (None, h2.stream.StreamState.CLOSED)


def apply_patch():
    monkey_patch_h2_reset_on_closed_stream_bug()
    monkey_patch_h2_data_to_send()



from purerpc.client import insecure_channel, secure_channel, Client
from purerpc.server import Service, Servicer, Server
from purerpc.rpc import Cardinality, RPCSignature, Stream
from purerpc.grpclib.status import Status, StatusCode
from purerpc.grpclib.exceptions import *
from purerpc.utils import run
from purerpc._version import __version__



import functools
from contextlib import AsyncExitStack

import anyio

from purerpc.grpc_proto import GRPCProtoSocket
from purerpc.grpclib.config import GRPCConfiguration
from purerpc.rpc import RPCSignature, Cardinality
from purerpc.wrappers import ClientStubUnaryUnary, ClientStubStreamStream, ClientStubUnaryStream, \
    ClientStubStreamUnary


class _Channel(AsyncExitStack):
    def __init__(self, host, port, ssl_context=None):
        super().__init__()
        self._host = host
        self._port = port
        self._ssl_context = ssl_context
        self._grpc_socket = None

    async def __aenter__(self):
        await super().__aenter__()  # Does nothing
        socket = await anyio.connect_tcp(self._host, self._port,
                                         ssl_context=self._ssl_context,
                                         tls=self._ssl_context is not None,
                                         tls_standard_compatible=False)
        config = GRPCConfiguration(client_side=True)
        self._grpc_socket = await self.enter_async_context(GRPCProtoSocket(config, socket))
        return self


def insecure_channel(host, port):
    return _Channel(host, port)

def secure_channel(host, port, ssl_context):
    return _Channel(host, port, ssl_context)


class Client:
    def __init__(self, service_name: str, channel: _Channel):
        self.service_name = service_name
        self.channel = channel

    async def rpc(self, method_name: str, request_type, response_type, metadata=None):
        message_type = request_type.DESCRIPTOR.full_name
        if metadata is None:
            metadata = ()
        stream = await self.channel._grpc_socket.start_request("http", self.service_name,
                                                               method_name, message_type,
                                                              "{}:{}".format(self.channel._host,
                                                                             self.channel._port),
                                                               custom_metadata=metadata)
        stream.expect_message_type(response_type)
        return stream

    def get_method_stub(self, method_name: str, signature: RPCSignature):
        stream_fn = functools.partial(self.rpc, method_name, signature.request_type,
                                      signature.response_type)
        if signature.cardinality == Cardinality.STREAM_STREAM:
            return ClientStubStreamStream(stream_fn)
        elif signature.cardinality == Cardinality.UNARY_STREAM:
            return ClientStubUnaryStream(stream_fn)
        elif signature.cardinality == Cardinality.STREAM_UNARY:
            return ClientStubStreamUnary(stream_fn)
        else:
            return ClientStubUnaryUnary(stream_fn)



class GRPCConfiguration:
    def __init__(self, client_side: bool, server_string=None, user_agent=None,
                 message_encoding=None, message_accept_encoding=None,
                 max_message_length=33554432):
        self._client_side = client_side
        if client_side and server_string is not None:
            raise ValueError("Passed client_side=True and server_string at the same time")
        if not client_side and user_agent is not None:
            raise ValueError("Passed user_agent put didn't pass client_side=True")
        self._server_string = server_string
        self._user_agent = user_agent
        self._max_message_length = max_message_length
        self._message_encoding = message_encoding

        # TODO: this does not need to be passed in config, may be just a single global string
        # with all encodings supported by grpclib
        if message_accept_encoding is not None:
            self._message_accept_encoding = ",".join(message_accept_encoding)
        else:
            self._message_accept_encoding = None

    @property
    def client_side(self):
        return self._client_side

    @property
    def server_string(self):
        return self._server_string

    @property
    def user_agent(self):
        return self._user_agent

    @property
    def message_encoding(self):
        return self._message_encoding

    @property
    def message_accept_encoding(self):
        return self._message_accept_encoding

    @property
    def max_message_length(self):
        return self._max_message_length



import re
import os
from setuptools import setup, find_packages

exec(open("src/purerpc/_version.py", encoding="utf-8").read())


def read(*names, **kwargs):
    with open(os.path.join(os.path.dirname(__file__), *names), "r") as fin:
        return fin.read()


def main():
    console_scripts = ['protoc-gen-purerpc=purerpc.protoc_plugin.plugin:main']
    setup(
        name="purerpc",
        version=__version__,
        license="Apache License Version 2.0",
        description=("Native, async Python gRPC client and server implementation "
                     "supporting asyncio, uvloop, trio"),
        long_description=(
            re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.md'))
        ),
        long_description_content_type='text/markdown',
        author="Andrew Stepanov",
        url="https://github.com/python-trio/purerpc",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Intended Audience :: Telecommunications Industry",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: MacOS",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX :: BSD",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: Implementation :: CPython",
            "Programming Language :: Python :: Implementation :: PyPy",
            "Framework :: AsyncIO",
            "Framework :: Trio",
            "Topic :: Internet",
            "Topic :: Internet :: WWW/HTTP",
            "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
            "Topic :: Software Development :: Libraries",
            "Topic :: Software Development :: Code Generators",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: System :: Networking",
        ],
        keywords=[
            "async", "await", "grpc", "pure-python", "pypy", "network",
            "rpc", "http2",
        ],
        packages=find_packages('src'),
        package_dir={'': 'src'},
        test_suite="tests",
        python_requires=">=3.7",
        install_requires=[
            "h2>=3.1.0,<4",
            "protobuf>=3.5.1",
            "anyio>=3.0.0",
            "tblib>=1.3.2",
            "async_generator>=1.10",  # for aclosing() only (Python < 3.10)
        ],
        entry_points={
            "console_scripts": console_scripts,
        },
        extras_require={
            'test': [
                "pytest",
                "grpcio>=1.25.0",           # fix version for PyPy
                "grpcio_tools>=1.25.0",     # same here
                "uvloop",
                "trio>=0.11",
                "python-forge>=18.6",
                "trustme",
            ]
        },
    )


if __name__ == "__main__":
    main()



import struct
import collections
from .exceptions import UnsupportedMessageEncodingError, MessageTooLargeError


class ByteBuffer:
    def __init__(self):
        self._deque = collections.deque()
        self._size = 0
        self._flow_controlled_length = 0

    def append(self, data, flow_controlled_length=None):
        if not isinstance(data, bytes) and not isinstance(data, bytearray):
            raise ValueError("Expected bytes")
        if flow_controlled_length is None:
            flow_controlled_length = len(data)
        if flow_controlled_length < len(data):
            raise ValueError("flow_controlled_length should be >= len(data)")
        self._deque.append((data, flow_controlled_length))
        self._size += len(data)

    def popleft_flowcontrol(self, amount):
        if amount > self._size:
            raise ValueError("Trying to extract {} bytes from ByteBuffer of length {}".format(
                amount, self._size))
        data = []
        flow_controlled_length = 0
        while amount > 0:
            next_element = self._deque[0][0]
            next_element_flow_controlled_length = self._deque[0][1]
            if amount >= len(next_element):
                self._size -= len(next_element)
                self._flow_controlled_length -= next_element_flow_controlled_length
                amount -= len(next_element)
                flow_controlled_length += next_element_flow_controlled_length
                data.append(next_element)
                self._deque.popleft()
            else:
                data.append(next_element[:amount])
                self._deque[0] = (next_element[amount:],
                                  next_element_flow_controlled_length - amount)
                self._size -= amount
                self._flow_controlled_length -= amount
                flow_controlled_length += amount
                amount = 0
        return b"".join(data), flow_controlled_length

    def popleft(self, amount):
        return self.popleft_flowcontrol(amount)[0]

    def __len__(self):
        return self._size

    @property
    def flow_controlled_length(self):
        return self._flow_controlled_length

    @property
    def length(self):
        return len(self)


class MessageReadBuffer:
    def __init__(self, message_encoding=None, max_message_length=4*1024*1024):
        self._message_encoding = message_encoding
        self._max_message_length = max_message_length

        self._buffer = ByteBuffer()
        self._messages = collections.deque()

        # MessageReadBuffer parser state
        self._compressed_flag = None
        self._message_length = None
        self._flow_controlled_length = None

    def data_received(self, data: bytes, flow_controlled_length=None):
        self._buffer.append(data, flow_controlled_length)
        self._process_new_messages()

    def decompress(self, data):
        if self._message_encoding == "gzip" or self._message_encoding == "deflate":
            import zlib
            return zlib.decompress(data)
        elif self._message_encoding == "snappy":
            import snappy
            return snappy.decompress(data)
        else:
            raise UnsupportedMessageEncodingError(
                "Unsupported compression: {}".format(self._message_encoding))

    def _parse_one_message(self):
        # either compressed_flag = message_length = flow_controlled_length = None and
        # compressed_flag, message_length are the next elements in self._buffer, or they are all
        # not None, and the next element in self._buffer is data
        if self._message_length is None:
            if len(self._buffer) < 5:
                return None, 0
            message_header, self._flow_controlled_length = self._buffer.popleft_flowcontrol(5)
            self._compressed_flag, self._message_length = struct.unpack('>?I', message_header)
            if self._message_length > self._max_message_length:
                # Even after the error is raised, the state is not corrupted, and parsing
                # can be safely resumed
                raise MessageTooLargeError(
                    "Received message larger than max: {message_length} > {max_message_length}".format(
                        message_length=self._message_length,
                        max_message_length=self._max_message_length,
                    )
                )
        if len(self._buffer) < self._message_length:
            return None, 0
        data, flow_controlled_length = self._buffer.popleft_flowcontrol(self._message_length)
        flow_controlled_length += self._flow_controlled_length
        if self._compressed_flag:
            data = self.decompress(data)
        self._compressed_flag = self._message_length = self._flow_controlled_length = None
        return data, flow_controlled_length

    def _process_new_messages(self):
        while True:
            result, flow_controlled_length = self._parse_one_message()
            if result is not None:
                self._messages.append((result, flow_controlled_length))
            else:
                return

    def __len__(self):
        return len(self._messages)

    def read_all_complete_messages(self):
        messages = [message for message, _ in self._messages]
        self._messages = collections.deque()
        return messages

    def read_all_complete_messages_flowcontrol(self):
        messages = self._messages
        self._messages = collections.deque()
        return messages

    def read_message(self):
        return self._messages.popleft()[0]

    def read_message_flowcontrol(self):
        return self._messages.popleft()


class MessageWriteBuffer:
    def __init__(self, message_encoding=None, max_message_length=4*1024*1024):
        self._buffer = ByteBuffer()
        self._message_encoding = message_encoding
        self._max_message_length = max_message_length

    def compress(self, data):
        if self._message_encoding == "gzip" or self._message_encoding == "deflate":
            import zlib
            return zlib.compress(data)
        elif self._message_encoding == "snappy":
            import snappy
            return snappy.compress(data)
        else:
            raise UnsupportedMessageEncodingError(
                "Unsupported compression: {}".format(self._message_encoding))

    def write_message(self, data: bytes, compress=False):
        if compress:
            data = self.compress(data)
        if len(data) > self._max_message_length:
            raise MessageTooLargeError(
                "Trying to send message larger than max: {message_length} > {max_message_length}".format(
                    message_length=len(data),
                    max_message_length=self._max_message_length,
                )
            )
        self._buffer.append(struct.pack('>?I', compress, len(data)))
        self._buffer.append(data)

    def data_to_send(self, amount):
        return self._buffer.popleft(amount)

    def __len__(self):
        return len(self._buffer)



import sys
import inspect
import collections
import functools

import logging
from contextlib import asynccontextmanager, AsyncExitStack

import anyio
import anyio.abc
from anyio import TASK_STATUS_IGNORED
from anyio.streams.tls import TLSListener

from .grpclib.events import RequestReceived
from .grpclib.status import Status, StatusCode
from .grpclib.exceptions import RpcFailedError
from .utils import run as purerpc_run
from purerpc.grpc_proto import GRPCProtoStream, GRPCProtoSocket
from purerpc.rpc import RPCSignature, Cardinality
from purerpc.wrappers import call_server_unary_unary, \
    call_server_unary_stream, call_server_stream_unary, call_server_stream_stream

from .grpclib.connection import GRPCConfiguration

log = logging.getLogger(__name__)

BoundRPCMethod = collections.namedtuple("BoundRPCMethod", ["method_fn", "signature"])


class Service:
    def __init__(self, name):
        self.name = name
        self.methods = {}

    def add_method(self, method_name: str, method_fn, rpc_signature: RPCSignature,
                   method_signature: inspect.Signature = None):
        if method_signature is None:
            method_signature = inspect.signature(method_fn)
        if len(method_signature.parameters) == 1:
            def method_fn_with_headers(arg, request):
                return method_fn(arg)
        elif len(method_signature.parameters) == 2:
            if list(method_signature.parameters.values())[1].name == "request":
                method_fn_with_headers = method_fn
            else:
                raise ValueError("Expected second parameter 'request'")
        else:
            raise ValueError("Expected method_fn to have exactly one or two parameters")
        self.methods[method_name] = BoundRPCMethod(method_fn_with_headers, rpc_signature)

    def rpc(self, method_name):
        def decorator(func):
            signature = inspect.signature(func)
            if signature.return_annotation == signature.empty:
                raise ValueError("Only annotated methods can be used with Service.rpc() decorator")
            if len(signature.parameters) not in (1, 2):
                raise ValueError("Only functions with one or two parameters can be used with "
                                 "Service.rpc() decorator")
            parameter = next(iter(signature.parameters.values()))
            if parameter.annotation == parameter.empty:
                raise ValueError("Only annotated methods can be used with Service.rpc() decorator")

            rpc_signature = RPCSignature.from_annotations(parameter.annotation,
                                                          signature.return_annotation)
            self.add_method(method_name, func, rpc_signature, method_signature=signature)
            return func

        return decorator


class Servicer:
    @property
    def service(self) -> Service:
        raise NotImplementedError()


@asynccontextmanager
async def _service_wrapper(service=None, setup_fn=None, teardown_fn=None):
    if setup_fn is not None:
        yield await setup_fn()
    else:
        yield service

    if teardown_fn is not None:
        await teardown_fn()


class Server:
    def __init__(self, port=50055, ssl_context=None):
        self.port = port
        self._ssl_context = ssl_context
        self.services = {}

    def add_service(self, service=None, context_manager=None, setup_fn=None, teardown_fn=None, name=None):
        if (service is not None) + (context_manager is not None) + (setup_fn is not None) != 1:
            raise ValueError("Precisely one of service, context_manager or setup_fn should be set")
        if name is None:
            if hasattr(service, "name"):
                name = service.name
            elif hasattr(context_manager, "name"):
                name = context_manager.name
            elif hasattr(setup_fn, "name"):
                name = setup_fn.name
            else:
                raise ValueError("Could not infer name, please provide 'name' argument to this function call "
                                 "or define 'name' attribute on service, context_manager or setup_fn")
        if service is not None:
            self.services[name] = _service_wrapper(service=service)
        elif context_manager is not None:
            self.services[name] = context_manager
        elif setup_fn is not None:
            self.services[name] = _service_wrapper(setup_fn=setup_fn, teardown_fn=teardown_fn)
        else:
            raise ValueError("Shouldn't have happened")

    async def serve_async(self, *, task_status=TASK_STATUS_IGNORED):
        """Run the grpc server

        The task_status protocol lets the caller know when the server is
        listening, and yields the port number (same given to Server constructor).
        """

        # TODO: resource usage warning
        async with AsyncExitStack() as stack:
            tcp_server = await anyio.create_tcp_listener(local_port=self.port, reuse_port=True)
            # read the resulting port, in case it was 0
            self.port = tcp_server.extra(anyio.abc.SocketAttribute.local_port)
            if self._ssl_context:
                tcp_server = TLSListener(tcp_server, self._ssl_context,
                                         standard_compatible=False)
            task_status.started(self.port)

            services_dict = {}
            for key, value in self.services.items():
                services_dict[key] = await stack.enter_async_context(value)

            await tcp_server.serve(ConnectionHandler(services_dict))

    def serve(self, backend=None):
        """
        DEPRECATED - use serve_async() instead

        This function runs an entire async event loop (there can only be one
        per thread), and there is no way to know when the server is ready for
        connections.
        """
        purerpc_run(self.serve_async, backend=backend)


class ConnectionHandler:
    RECEIVE_BUFFER_SIZE = 65536

    def __init__(self, services: dict):
        self.config = GRPCConfiguration(client_side=False)
        self.grpc_socket = None
        self.services = services

    async def request_received(self, stream: GRPCProtoStream):
        try:
            await stream.start_response()
            event = await stream.receive_event()

            if not isinstance(event, RequestReceived):
                await stream.close(Status(StatusCode.INTERNAL, status_message="Expected headers"))
                return

            try:
                service = self.services[event.service_name]
            except KeyError:
                await stream.close(Status(
                    StatusCode.UNIMPLEMENTED,
                    status_message="Service {service_name} is not implemented".format(service_name=event.service_name)
                ))
                return

            try:
                bound_rpc_method = service.methods[event.method_name]
            except KeyError:
                await stream.close(Status(
                    StatusCode.UNIMPLEMENTED,
                    status_message="Method {method_name} is not implemented in service {service_name}".format(
                        method_name=event.method_name,
                        service_name=event.service_name
                    )
                ))
                return

            # TODO: Should at least pass through GeneratorExit
            try:
                method_fn = functools.partial(bound_rpc_method.method_fn, request=event)
                cardinality = bound_rpc_method.signature.cardinality
                stream.expect_message_type(bound_rpc_method.signature.request_type)
                if cardinality == Cardinality.STREAM_STREAM:
                    await call_server_stream_stream(method_fn, stream)
                elif cardinality == Cardinality.UNARY_STREAM:
                    await call_server_unary_stream(method_fn, stream)
                elif cardinality == Cardinality.STREAM_UNARY:
                    await call_server_stream_unary(method_fn, stream)
                else:
                    await call_server_unary_unary(method_fn, stream)
            except RpcFailedError as error:
                await stream.close(error.status)
            except:
                # TODO: limit catch to Exception, so async cancel can propagate
                log.warning("Got exception while writing response stream",
                            exc_info=log.getEffectiveLevel() == logging.DEBUG)
                await stream.close(Status(StatusCode.CANCELLED, status_message=repr(sys.exc_info())))
        except:
            # TODO: limit catch to Exception, so async cancel can propagate
            log.warning("Got exception in request_received",
                        exc_info=log.getEffectiveLevel() == logging.DEBUG)

    async def __call__(self, stream_: anyio.abc.SocketStream):
        # TODO: Should at least pass through GeneratorExit
        try:
            async with GRPCProtoSocket(self.config, stream_) as self.grpc_socket:
                # TODO: resource usage warning
                # TODO: TaskGroup() uses a lot of memory if the connection is kept for a long time
                # TODO: do we really need it here?
                async with anyio.create_task_group() as task_group:
                    async for stream in self.grpc_socket.listen():
                        task_group.start_soon(self.request_received, stream)
        except:
            # TODO: limit catch to Exception, so async cancel can propagate
            log.warning("Got exception in main dispatch loop",
                        exc_info=log.getEffectiveLevel() == logging.DEBUG)



import sys
import enum
import socket
import datetime
from contextlib import AsyncExitStack
from typing import Dict

import anyio
import anyio.abc
from purerpc.utils import is_darwin, is_windows
from purerpc.grpclib.exceptions import ProtocolError

from .grpclib.connection import GRPCConfiguration, GRPCConnection
from .grpclib.events import RequestReceived, RequestEnded, ResponseEnded, MessageReceived, WindowUpdated
from .grpclib.buffers import MessageWriteBuffer
from .grpclib.exceptions import StreamClosedError


class SocketWrapper(AsyncExitStack):
    def __init__(self, grpc_connection: GRPCConnection, stream: anyio.abc.SocketStream):
        super().__init__()
        self._set_socket_options(stream)
        self._stream = stream
        self._grpc_connection = grpc_connection
        self._flush_event = anyio.Event()
        self._running = True

    async def __aenter__(self):
        await super().__aenter__()
        task_group = await self.enter_async_context(anyio.create_task_group())
        task_group.start_soon(self._writer_thread)

        async def callback():
            self._running = False
            self._flush_event.set()

        self.push_async_callback(callback)
        return self

    @staticmethod
    def _set_socket_options(stream: anyio.abc.SocketStream):
        sock = stream.extra(anyio.abc.SocketAttribute.raw_socket)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if hasattr(socket, "TCP_KEEPIDLE"):
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 300)
        elif is_darwin():
            # Darwin specific option
            TCP_KEEPALIVE = 16
            sock.setsockopt(socket.IPPROTO_TCP, TCP_KEEPALIVE, 300)
        if not is_windows():
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 30)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    async def _writer_thread(self):
        while True:
            data = self._grpc_connection.data_to_send()
            if data:
                await self._stream.send(data)
            elif self._running:
                await self._flush_event.wait()
                self._flush_event = anyio.Event()
            else:
                return

    async def flush(self):
        """This maybe called from different threads."""
        self._flush_event.set()

    async def recv(self, buffer_size: int):
        """This may only be called from single thread."""
        return await self._stream.receive(buffer_size)


class GRPCStreamState(enum.Enum):
    OPEN = 1
    HALF_CLOSED_REMOTE = 2
    HALF_CLOSED_LOCAL = 3
    CLOSED = 4


class GRPCStream:
    def __init__(self, grpc_connection: GRPCConnection, stream_id: int, socket: SocketWrapper,
                 grpc_socket: "GRPCSocket"):
        self._stream_id = stream_id
        self._grpc_connection = grpc_connection
        self._grpc_socket = grpc_socket
        self._socket = socket
        self._flow_control_update_event = anyio.Event()
        # TODO: find a reasonable buffer size, or expose it in the API
        self._incoming_events = anyio.create_memory_object_stream(max_buffer_size=sys.maxsize)  # (send, receive)
        self._response_started = False
        self._state = GRPCStreamState.OPEN
        self._start_stream_event = None
        self._end_stream_event = None

    @property
    def state(self):
        return self._state

    @property
    def start_stream_event(self):
        return self._start_stream_event

    @property
    def end_stream_event(self):
        return self._end_stream_event

    @property
    def stream_id(self):
        return self._stream_id

    @property
    def client_side(self):
        return self._grpc_connection.config.client_side

    @property
    def debug_prefix(self):
        return "[CLIENT] " if self.client_side else "[SERVER] "

    def _close_remote(self):
        if self._state == GRPCStreamState.OPEN:
            self._state = GRPCStreamState.HALF_CLOSED_REMOTE
        elif self._state == GRPCStreamState.HALF_CLOSED_LOCAL:
            self._state = GRPCStreamState.CLOSED
            del self._grpc_socket._streams[self._stream_id]

    def _close_local(self):
        if self._state == GRPCStreamState.OPEN:
            self._state = GRPCStreamState.HALF_CLOSED_LOCAL
        elif self._state == GRPCStreamState.HALF_CLOSED_REMOTE:
            self._state = GRPCStreamState.CLOSED
            del self._grpc_socket._streams[self._stream_id]

    async def _set_flow_control_update(self):
        self._flow_control_update_event.set()

    async def _wait_flow_control_update(self):
        await self._flow_control_update_event.wait()
        self._flow_control_update_event = anyio.Event()

    async def _send(self, message: bytes, compress=False):
        message_write_buffer = MessageWriteBuffer(self._grpc_connection.config.message_encoding,
                                                  self._grpc_connection.config.max_message_length)
        message_write_buffer.write_message(message, compress)
        while message_write_buffer:
            window_size = self._grpc_connection.flow_control_window(self._stream_id)
            if window_size <= 0:
                await self._wait_flow_control_update()
                continue
            num_data_to_send = min(window_size, len(message_write_buffer))
            data = message_write_buffer.data_to_send(num_data_to_send)
            self._grpc_connection.send_data(self._stream_id, data)
            await self._socket.flush()

    async def _receive(self):
        event = await self._incoming_events[1].receive()
        if isinstance(event, MessageReceived):
            self._grpc_connection.acknowledge_received_data(self._stream_id,
                                                            event.flow_controlled_length)
            await self._socket.flush()
        elif isinstance(event, RequestEnded) or isinstance(event, ResponseEnded):
            assert self._end_stream_event is None
            self._end_stream_event = event
        else:
            assert self._start_stream_event is None
            self._start_stream_event = event
        return event

    async def close(self, status=None, content_type_suffix="", custom_metadata=()):
        if self.client_side and (status or custom_metadata):
            raise ValueError("Client side streams cannot be closed with non-default arguments")
        if self._state in (GRPCStreamState.HALF_CLOSED_LOCAL, GRPCStreamState.CLOSED):
            raise TypeError("Closing already closed stream")
        self._close_local()
        if self.client_side:
            try:
                self._grpc_connection.end_request(self._stream_id)
            except StreamClosedError:
                # Remote end already closed connection, do nothing here
                pass
        elif self._response_started:
            self._grpc_connection.end_response(self._stream_id, status, custom_metadata)
        else:
            self._grpc_connection.respond_status(self._stream_id, status,
                                                 content_type_suffix, custom_metadata)
        await self._socket.flush()

    async def start_response(self, content_type_suffix="", custom_metadata=()):
        if self.client_side:
            raise ValueError("Cannot start response on client-side socket")
        self._grpc_connection.start_response(self._stream_id, content_type_suffix, custom_metadata)
        self._response_started = True
        await self._socket.flush()


# TODO: this name is not correct, should be something like GRPCConnection (but this name is already
# occupied)
class GRPCSocket(AsyncExitStack):
    StreamClass = GRPCStream

    def __init__(self, config: GRPCConfiguration, sock,
                 receive_buffer_size=1024*1024):
        super().__init__()
        self._grpc_connection = GRPCConnection(config=config)
        self._socket = SocketWrapper(self._grpc_connection, sock)
        self._receive_buffer_size = receive_buffer_size
        self._streams = {}  # type: Dict[int, GRPCStream]

    async def __aenter__(self):
        await super().__aenter__()
        self._socket = await self.enter_async_context(self._socket)
        self._grpc_connection.initiate_connection()
        await self._socket.flush()
        if self.client_side:
            task_group = await self.enter_async_context(anyio.create_task_group())
            self.callback(task_group.cancel_scope.cancel)
            task_group.start_soon(self._reader_thread)
        return self

    @property
    def client_side(self):
        return self._grpc_connection.config.client_side

    def _stream_ctor(self, stream_id):
        return self.StreamClass(self._grpc_connection, stream_id, self._socket, self)

    def _allocate_stream(self, stream_id):
        self._streams[stream_id] = self._stream_ctor(stream_id)
        return self._streams[stream_id]

    async def _listen(self):
        while True:
            data = await self._socket.recv(self._receive_buffer_size)
            if not data:
                return
            events = self._grpc_connection.receive_data(data)
            await self._socket.flush()
            for event in events:
                if isinstance(event, WindowUpdated):
                    if event.stream_id == 0:
                        for stream in self._streams.values():
                            await stream._set_flow_control_update()
                    elif event.stream_id in self._streams:
                        await self._streams[event.stream_id]._set_flow_control_update()
                    continue
                elif isinstance(event, RequestReceived):
                    self._allocate_stream(event.stream_id)

                await self._streams[event.stream_id]._incoming_events[0].send(event)

                if isinstance(event, RequestReceived):
                    yield self._streams[event.stream_id]
                elif isinstance(event, ResponseEnded) or isinstance(event, RequestEnded):
                    self._streams[event.stream_id]._close_remote()

    async def _reader_thread(self):
        async for _ in self._listen():
            raise ProtocolError("Received request on client end")

    async def listen(self):
        if self.client_side:
            raise ValueError("Cannot listen client-side socket")

        async for value in self._listen():
            yield value

    async def start_request(self, scheme: str, service_name: str, method_name: str,
                            message_type=None, authority=None, timeout: datetime.timedelta=None,
                            content_type_suffix="", custom_metadata=()):
        if not self.client_side:
            raise ValueError("Cannot start request on server-side socket")
        stream_id = self._grpc_connection.get_next_available_stream_id()
        stream = self._allocate_stream(stream_id)
        self._grpc_connection.start_request(stream_id, scheme, service_name, method_name,
                                            message_type, authority, timeout,
                                            content_type_suffix, custom_metadata)
        await self._socket.flush()
        return stream



import functools
import collections
import subprocess
import multiprocessing
import tempfile
import shutil
import os
import sys
import inspect
import importlib
import concurrent.futures
import contextlib
import time
import random
import string
from multiprocessing.connection import Connection

from tblib import pickling_support
pickling_support.install()

import forge
import anyio
from async_generator import aclosing

from .utils import run as purerpc_run

# work around pickle issue on macOS
if sys.platform == 'darwin':
    multiprocessing = multiprocessing.get_context('fork')


@contextlib.contextmanager
def compile_temp_proto(*relative_proto_paths):
    modules = []
    with tempfile.TemporaryDirectory() as temp_dir:
        sys.path.insert(0, temp_dir)
        try:
            for relative_proto_path in relative_proto_paths:
                proto_path = os.path.join(os.path.dirname(
                    inspect.currentframe().f_back.f_back.f_globals['__file__']),
                    relative_proto_path)
                proto_filename = os.path.basename(proto_path)
                proto_temp_path = os.path.join(temp_dir, proto_filename)
                shutil.copyfile(proto_path, proto_temp_path)
            for relative_proto_path in relative_proto_paths:
                proto_filename = os.path.basename(relative_proto_path)
                proto_temp_path = os.path.join(temp_dir, proto_filename)
                cmdline = [sys.executable, '-m', 'grpc_tools.protoc',
                           '--python_out=.', '--purerpc_out=.', '--grpc_python_out=.',
                           '-I' + temp_dir, proto_temp_path]
                subprocess.check_call(cmdline, cwd=temp_dir)

                pb2_module_name = proto_filename.replace(".proto", "_pb2")
                pb2_grpc_module_name = proto_filename.replace(".proto", "_pb2_grpc")
                grpc_module_name = proto_filename.replace(".proto", "_grpc")

                pb2_module = importlib.import_module(pb2_module_name)
                pb2_grpc_module = importlib.import_module(pb2_grpc_module_name)
                grpc_module = importlib.import_module(grpc_module_name)
                modules.extend((pb2_module, pb2_grpc_module, grpc_module))
            yield modules
        finally:
            sys.path.remove(temp_dir)


_WrappedResult = collections.namedtuple("_WrappedResult", ("result", "exc_info"))


def _wrap_gen_in_process(conn: Connection):
    def decorator(gen):
        @functools.wraps(gen)
        def new_func(*args, **kwargs):
            try:
                for elem in gen(*args, **kwargs):
                    conn.send(_WrappedResult(result=elem, exc_info=None))
            except:
                conn.send(_WrappedResult(result=None, exc_info=sys.exc_info()))
            finally:
                conn.close()
        return new_func
    return decorator


async def async_iterable_to_list(async_iterable):
    result = []
    async with aclosing(async_iterable) as async_iterable:
        async for value in async_iterable:
            result.append(value)
    return result


def random_payload(min_size=1000, max_size=100000):
    return "".join(random.choice(string.ascii_letters)
                   for _ in range(random.randint(min_size, max_size)))


@contextlib.contextmanager
def _run_context_manager_generator_in_process(cm_gen):
    parent_conn, child_conn = multiprocessing.Pipe(duplex=False)
    target_fn = _wrap_gen_in_process(child_conn)(cm_gen)

    process = multiprocessing.Process(target=target_fn)
    process.start()
    try:
        wrapped_result = parent_conn.recv()
        if wrapped_result.exc_info is not None:
            raise wrapped_result.exc_info[0].with_traceback(*wrapped_result.exc_info[1:])
        else:
            yield wrapped_result.result
    finally:
        try:
            if parent_conn.poll():
                exc_info = parent_conn.recv().exc_info
                if exc_info is not None:
                    raise exc_info[0].with_traceback(*exc_info[1:])
        finally:
            process.terminate()
            process.join()
            parent_conn.close()


def run_purerpc_service_in_process(service, ssl_context=None):
    # TODO: there is no reason to run the server as a separate process...
    #   just use serve_async().  This synchronous cm has timing problems,
    #   because the server may not be listening before yielding to the body.

    def target_fn():
        import purerpc
        import socket

        # Grab an ephemeral port in advance, because we need to yield the port
        # before blocking on serve()...
        with socket.socket() as sock:
            sock.bind(('127.0.0.1', 0))
            port = sock.getsockname()[1]

        server = purerpc.Server(port=port, ssl_context=ssl_context)
        server.add_service(service)
        yield port
        server.serve()

        # async def sleep_10_seconds_then_die():
        #     await anyio.sleep(20)
        #     raise ValueError
        #
        # async def main():
        #     async with anyio.create_task_group() as tg:
        #         tg.start_soon(server.serve_async)
        #         tg.start_soon(sleep_10_seconds_then_die)
        #
        # import cProfile
        # cProfile.runctx("purerpc_run(main)", globals(), locals(), sort="tottime")

    return _run_context_manager_generator_in_process(target_fn)


# TODO: remove grpcio dependency from tests.  There is no reason to unit test
#  grpc project's code, and it's blocking pypy support.
def run_grpc_service_in_process(add_handler_fn):
    def target_fn():
        import grpc
        server = grpc.server(concurrent.futures.ThreadPoolExecutor(max_workers=1))
        port = server.add_insecure_port('[::]:0')
        add_handler_fn(server)
        server.start()
        yield port
        while True:
            time.sleep(60)
    return _run_context_manager_generator_in_process(target_fn)


def run_tests_in_workers(*, target, num_workers):
    parent_conn, child_conn = multiprocessing.Pipe(duplex=False)

    @_wrap_gen_in_process(child_conn)
    def target_fn():
        target()
        yield

    processes = [multiprocessing.Process(target=target_fn) for _ in range(num_workers)]
    for process in processes:
        process.start()

    try:
        for _ in range(num_workers):
            wrapped_result = parent_conn.recv()
            if wrapped_result.exc_info is not None:
                raise wrapped_result.exc_info[0].with_traceback(*wrapped_result.exc_info[1:])
    finally:
        parent_conn.close()
        for process in processes:
            process.join()


def async_test(corofunc):
    if not inspect.iscoroutinefunction(corofunc):
        raise TypeError("Expected coroutine function")

    @functools.wraps(corofunc)
    def func(**kwargs):
        return purerpc_run(functools.partial(corofunc, **kwargs))
    return func


def grpc_client_parallelize(num_workers):
    def decorator(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            def target():
                func(*args, **kwargs)
            run_tests_in_workers(target=target, num_workers=num_workers)

        new_func.__parallelized__ = True
        return new_func
    return decorator


def purerpc_client_parallelize(num_tasks):
    def decorator(corofunc):
        if not inspect.iscoroutinefunction(corofunc):
            raise TypeError("Expected coroutine function")

        @functools.wraps(corofunc)
        async def new_corofunc(**kwargs):
            async with anyio.create_task_group() as tg:
                for _ in range(num_tasks):
                    tg.start_soon(functools.partial(corofunc, **kwargs))
        return new_corofunc
    return decorator


def grpc_channel(port_fixture_name, channel_arg_name="channel"):
    def decorator(func):
        if hasattr(func, "__parallelized__") and func.__parallelized__:
            raise TypeError("Cannot pass gRPC channel to already parallelized test, grpc_client_parallelize should "
                            "be the last decorator in chain")

        @forge.compose(
            forge.copy(func),
            forge.modify(channel_arg_name, name=port_fixture_name, interface_name="port_fixture_value"),
        )
        def new_func(*, port_fixture_value, **kwargs):
            import grpc
            with grpc.insecure_channel('127.0.0.1:{}'.format(port_fixture_value)) as channel:
                func(**kwargs, channel=channel)

        return new_func
    return decorator


def purerpc_channel(port_fixture_name, channel_arg_name="channel"):
    def decorator(corofunc):
        if not inspect.iscoroutinefunction(corofunc):
            raise TypeError("Expected coroutine function")

        @forge.compose(
            forge.copy(corofunc),
            forge.modify(channel_arg_name, name=port_fixture_name, interface_name="port_fixture_value"),
        )
        async def new_corofunc(*, port_fixture_value, **kwargs):
            import purerpc
            async with purerpc.insecure_channel("127.0.0.1", port_fixture_value) as channel:
                await corofunc(**kwargs, channel=channel)

        return new_corofunc
    return decorator
